from .find_first import find_first
from noodles import (gather, schedule, unpack)
from typing import (Any, Callable, Iterable)


@schedule
def all(pred: Callable, xs: Iterable):
    """
    Check whether all the elements of the iterable `xs`
    fullfill predicate `pred`.

    :param pred:
       predicate function
    :param xs:
       iterable object.
    :returns: boolean
    """
    for x in xs:
        if not pred(x):
            return False

    return True


@schedule
def any(pred: Callable, xs: Iterable):
    """
    Check if at least one element of the iterable `xs`
    fullfills predicate `pred`.

    :param pred:
       predicate function.
    :param xs:
       iterable object.
    :returns: boolean
    """
    b = find_first(pred, xs)

    return True if b is not None else False


@schedule
def filter(pred: Callable, xs: Iterable):
    """
    Applied a predicate to a list returning a :py:class:`PromisedObject`
    containing the values satisfying the predicate.

    :param pred: predicate function.
    :param xs: iterable object.
    :returns: :py:class:`PromisedObject`
    """
    generator = (x for x in xs if pred(x))

    return gather(*generator)


@schedule
def fold(
        fun: Callable, state: Any, xs: Iterable):
    """
    Traverse an iterable object while performing stateful computations
    with the elements. It returns a :py:class:`PromisedObject` containing
    the result of the stateful computations.

    For a general definition of folding see:
    https://en.wikipedia.org/wiki/Fold_(higher-order_function)

    :param fun: stateful function.
    :param state: initial state.
    :param xs: iterable object.
    :returns: :py:class:`PromisedObject`
    """
    def generator(state):
        for x in xs:
            state, r = unpack(fun(state, x), 2)
            yield r

    return gather(*generator(state))


@schedule
def map(fun: Callable, xs: Iterable):
    """
    Traverse an iterable object applying function `fun`
    to each element and finally creats a workflow from it.

    :param fun:
      function to call in each element of the iterable
      object.
    :param xs:
      Iterable object.

    returns::py:class:`PromisedObject`
    """
    generator = (fun(x) for x in xs)

    return gather(*generator)


@schedule
def zip_with(fun: Callable, xs: Iterable, ys: Iterable):
    """
    Fuse two Iterable object using the function `fun`.
    Notice that if the two objects have different len,
    the shortest object gives the result's shape.

    :param fun:
       function taking two argument use to process
       element x from `xs` and y from `ys`.

    :param xs:
       first iterable.

    :param ys:
       second iterable.

    returns::py:class:`PromisedObject`
    """
    generator = (fun(*rs) for rs in zip(xs, ys))

    return gather(*generator)



import uuid
import xenon

from ..remote.worker_config import WorkerConfig


class XenonJobConfig(WorkerConfig):
    def __init__(
            self, *, queue_name=None, environment=None, options=None,
            time_out=1000, scheduler_arguments=None, **kwargs):
        super(XenonJobConfig, self).__init__(**kwargs)
        self.time_out = time_out

        executable, arguments = self.command_line()
        self.xenon_job_description = xenon.JobDescription(
            executable=str(executable),
            arguments=arguments,
            working_directory=str(self.working_dir),
            queue_name=queue_name,
            environment=environment,
            scheduler_arguments=scheduler_arguments,
            options=options)


class Machine(object):
    """Configuration to the Xenon library.

    Xenon is a Java library that offers a uniform interface to execute jobs.
    These jobs may be run locally, over ssh ar against a queue manager like
    SLURM.

    [Documentation to PyXenon can be found online](http://pyxenon.rtfd.io/)

    :param name:
        The quasi human readable name to give to this Xenon instance.
        This defaults to a generated UUID.

    :param jobs_scheme:
        The scheme by which to schedule jobs. Should be one of 'local', 'ssh',
        'slurm' etc. See the Xenon documentation.

    :param files_scheme:
        The scheme by which to transfer files. Should be 'local' or 'ssh'.
        See the Xenon documentation.

    :param location:
        A location. This can be the host of the 'ssh' or 'slurm' server.

    :param credential:
        To enter a server through ssh, we need to have some credentials.
        Preferably, you have a private/public key pair by which you can
        identify yourself. Otherwise, this would be a combination of
        username/password. This functions that can create a credential
        object can be found in Xenon.credentials in the Xenon documentation.

    :param jobs_properties:
        Configuration to the Xenon.jobs module.

    :param files_properties:
        Configuration to the Xenon.files module.
    """
    def __init__(self, *, name=None, scheduler_adaptor='local',
                 location=None, credential=None, jobs_properties=None,
                 files_properties=None):
        self.name = name or ("xenon-" + str(uuid.uuid4()))
        self.scheduler_adaptor = scheduler_adaptor
        self.location = location
        self.credential = credential
        self.jobs_properties = jobs_properties
        self.files_properties = files_properties
        self._scheduler = None
        self._file_system = None

    @property
    def scheduler_args(self):
        args = {'adaptor': self.scheduler_adaptor,
                'location': self.location,
                'properties': self.jobs_properties}

        if isinstance(self.credential, xenon.PasswordCredential):
            args['password_credential'] = self.credential
        if isinstance(self.credential, xenon.CertificateCredential):
            args['certificate_credential'] = self.credential

        return args

    @property
    def scheduler(self):
        """Returns the scheduler object."""
        if self._scheduler is None:
            self._scheduler = xenon.Scheduler.create(**self.scheduler_args)

        return self._scheduler

    @property
    def file_system(self):
        """Gets the filesystem corresponding to the open scheduler."""
        if self._file_system is None:
            self._file_system = self.scheduler.get_file_system()

        return self._file_system



"""
Coroutine streaming module
==========================

.. note::
    In a break with tradition, some classes in this module have lower case
    names because they tend to be used as function decorators.

We use coroutines to communicate messages between different components
in the Noodles runtime. Coroutines can have input or output in two ways
*passive* and *active*. An example:

.. code-block:: python

    def f_pulls(coroutine):
        for msg in coroutine:
            print(msg)

    def g_produces(lines):
        for l in lines:
            yield lines

    lines = ['aap', 'noot', 'mies']

    f_pulls(g_produces(lines))

This prints the words 'aap', 'noot' and 'mies'. This same program could be
written where the co-routine is the one receiving messages:

.. code-block:: python

    def f_receives():
        while True:
            msg = yield
            print(msg)

    def g_pushes(coroutine, lines):
        for l in lines:
            coroutine.send(l)

    sink = f_receives()
    sink.send(None)  # the co-routine needs to be initialised
                     # alternatively, .next() does the same as .send(None)
    g_pushes(sink, lines)

The action of creating a coroutine and setting it to the first `yield`
statement can be performed by a little decorator:

.. code-block:: python

    from functools import wraps

    def coroutine(f):
        @wraps(f)
        def g(*args, **kwargs):
            sink = f(*args, **kwargs)
            sink.send(None)
            return sink

        return g

Pull and push
-------------

The |pull| and |push| classes capture the idea of pushing and pulling
coroutines, wrapping them in an object. These objects can then be chained
using the ``>>`` operator. Example:

.. code-block:: python

    >>> from noodles.lib import (pull_map, pull_from)
    >>> @pull_map
    ... def square(x):
    ...     return x*x
    ...
    >>> squares = pull_from(range(10)) >> square
    >>> list(squares)
    [0, 1, 4, 9, 16, 25, 36, 49, 64, 81]

Queues
------

Queues in python are thread-safe objects. We can define a new |Queue| object
that uses the python `queue.Queue` to buffer and distribute messages over
several threads:

.. code-block:: python

    import queue

    class Queue(object):
        def __init__(self):
            self._q = queue.Queue()

        def source(self):
            while True:
                msg = self._q.get()
                yield msg
                self._q.task_done()

        @coroutine
        def sink(self):
            while True:
                msg = yield
                self._q.put(msg)

        def wait(self):
            self.Q.join()

Note, that both ends of the queue are, as we call it, passive. We could make
an active source (it would become a normal function), taking a call-back as
an argument. However, we're designing the Noodles runtime so that it easy to
interleave functionality. Moreover, the `Queue` object is only concerned
with the state of its own queue. The outside universe is only represented by
the `yield` statements, thus preserving the principle of encapsulation.
"""

from .decorator import (
    decorator)

from .coroutine import (
    coroutine)

from .streams import (
    stream, pull, push, pull_map, push_map, sink_map,
    broadcast, branch, patch, pull_from, push_from)

from .connection import (
    Connection)

from .queue import (
    Queue, EndOfQueue, FlushQueue)

from .thread_pool import (
    thread_counter, thread_pool)

from .utility import (
    object_name, look_up, importable, deep_map, inverse_deep_map, unwrap,
    is_unwrapped)

__all__ = [
    'decorator', 'coroutine',
    'stream', 'pull', 'push', 'pull_map', 'push_map', 'sink_map',
    'broadcast', 'branch', 'patch', 'pull_from', 'push_from',
    'Connection', 'Queue', 'EndOfQueue', 'FlushQueue',
    'thread_pool', 'thread_counter',
    'object_name', 'look_up', 'importable', 'deep_map', 'inverse_deep_map',
    'unwrap', 'is_unwrapped']



"""
Test the use of init and finish functions.
"""

from noodles import schedule, run_process, serial


def init():
    """Creates a global variable ``s``, and returns True."""
    global S
    S = "This global variable needs to be here!"
    return True


def finish():
    """Just print a message."""
    return "Finish functino was run!"


@schedule
def checker():
    """Check if global variable ``s`` exists."""
    return S == "This global variable needs to be here!"


def test_globals():
    """Test init and finish functions on ``run_process``."""
    a = checker()
    result = run_process(a, n_processes=1, registry=serial.base,
                         init=init, finish=finish)
    assert result



"""
SQLite3 job database
--------------------

The database contains three tables:
    - jobs
    - sessions
    - timestamps

Each job that is started will get an entry in the "jobs" table.
    1. If a job with the same "prov" field exists and it has a result, we mark
    the new job by identifying the existing result in the "link" field.
    2. If a job with the same "prov" field exists, but has no result and has
    the same "session" as the new job, the new job is marked link of the
    older one. We keep a note in memory that, when the result is known, it is
    also given for the newer job.
    3. Otherwise, we pass the job on to a worker and wait for the result.

In the case that a job returns a new workflow, we'd like to identify the result
of that workflow with that of the original job. In that case we add a "link"
once a non-workflow result is known.
"""

import sqlite3
from threading import Lock
from collections import (defaultdict)
# from typing import NamedTuple
from collections import (namedtuple)
# import datetime
import sys
from enum import IntEnum

from ..run.messages import (JobMessage, ResultMessage)
from ..workflow import (is_workflow, get_workflow, FunctionNode, NodeData)
from .key import (prov_key)

try:
    import ujson as json
except ImportError:
    import json


schema = '''
    create table if not exists "jobs" (
        "id"          integer unique primary key,
        "session"     integer not null references "sessions"("id")
                      on delete cascade,
        "name"        text,
        "status"      integer references "status"("id"),
        "prov"        text,
        "version"     text,
        "function"    text,
        "arguments"   text,
        "result"      text,
        "link"        integer references "jobs"("id") );

    create index if not exists "prov" on "jobs"("prov");

    create table if not exists "sessions" (
        "id"        integer unique primary key,
        "time"      datetime default current_timestamp,
        "info"      text );

    create table if not exists "timestamps" (
        "job"       integer not null references "jobs"("id")
                    on delete cascade,
        "time"      datetime default current_timestamp,
        "what"      text );

    create table if not exists "status" (
        "id"        integer unique primary key,
        "name"      text );

    insert or ignore into "status" values
        (0, "INACTIVE"),    -- just barely known about
        (1, "WAITING"),     -- fully registered but no result yet
        (2, "STORED"),      -- the result of the job is stored
        (3, "WORKFLOW"),    -- the result of the job was a workflow
        (4, "DUPLICATE"),   -- the job was a duplicate
        (5, "LINKEE");      -- the job was not registered, but its result is
                            -- stored, because a linked job is registered
'''


JobEntry = namedtuple('JobEntry', [
    'id', 'session', 'name', 'status', 'prov', 'version', 'function',
    'arguments', 'result', 'link'])


# class JobEntry(NamedTuple):
#     """Python tuple reflection of Job entry in database."""
#     id: int
#     session: int
#     name: str
#     status: int
#     prov: str
#     version: str
#     function: str
#     arguments: str
#     result: str
#     link: int

SessionEntry = namedtuple('SessionEntry', [
    'id', 'time', 'info'])

# class SessionEntry(NamedTuple):
#     """Python tuple reflection of Session entry in database."""
#     id: int
#     time: datetime.datetime
#     info: str

TimestampEntry = namedtuple('TimestampEntry', [
    'job', 'time', 'what'])

# class TimestampEntry(NamedTuple):
#     """Python tuple reflection of Timestamp entry in database."""
#     job: int
#     time: datetime.datetime
#     what: str


class Status(IntEnum):
    INACTIVE = 0
    WAITING = 1
    STORED = 2
    WORKFLOW = 3
    DUPLICATE = 4
    LINKEE = 5


class JobDB:
    """Keeps a database of jobs, with a MD5 hash that encodes the function
    name, version, and all arguments to the function.
    """
    def __init__(self, path, registry, info=None):
        self.attached = defaultdict(list)
        self.connection = sqlite3.connect(path, check_same_thread=False)
        self.jobs = {}
        self.links = defaultdict(list)
        self.registry = registry()

        self.cur = self.connection.cursor()
        self.cur.executescript(schema)
        self.lock = Lock()
        self.workflows = {}

        with self.lock:
            self.cur.execute(
                'insert into "sessions" ("info") values (?)', (info,))
            self.session = self.cur.lastrowid

    def __enter__(self):
        return self

    def __exit__(self, exc, exc_type, stacktrace):
        self.connection.commit()
        self.connection.close()

    # --------- job-keeper interface ------------
    def __len__(self):
        with self.lock:
            return len(self.jobs)

    def __delitem__(self, key):
        with self.lock:
            del self.jobs[key]

    def __getitem__(self, key):
        with self.lock:
            return self.jobs[key]

    def register(self, job):
        """Takes a job (unencoded) and adorns it with a unique key; this makes
        an entry in the database without any further specification."""
        with self.lock:
            self.cur.execute(
                'insert into "jobs" ("name", "session", "status") '
                'values (?, ?, ?)', (job.name, self.session, Status.INACTIVE))
            self.jobs[self.cur.lastrowid] = job
            return JobMessage(self.cur.lastrowid, job.node)

    def store_result(self, key, status, value, _):
        """Store the result of a job back in the node; this does nothing to the
        database."""
        if status != 'done':
            return

        if key not in self.jobs:
            print("WARNING: store_result called but job not in registry:\n"
                  "   race condition? Not doing anything.\n", file=sys.stderr)
            return

        with self.lock:
            job = self.jobs[key]
            job.node.result = value

    # --------- database interface ---------------------
    def list_jobs(self):
        with self.lock:
            self.cur.execute(
                'select "id", "function", "arguments" from "jobs"')
            return {
                x[0]: FunctionNode.from_node_data(NodeData(
                    self.registry.from_json(x[1]),
                    self.registry.from_json(x[2]),
                    {}
                )) for x in self.cur.fetchall()}

    def get_result(self, db_id):
        with self.lock:
            self.cur.execute(
                'select * from "jobs" where "id" = ?',
                (db_id,))
            rec = self.cur.fetchone()
            if rec is None:
                raise ValueError("No record found with id %s", db_id)
            rec = JobEntry(*rec)

            if rec.result is not None and rec.status == Status.WORKFLOW:
                # the found duplicate returned a workflow
                if rec.link is not None:
                    # link is set, so result is fully realized
                    self.cur.execute(
                        'select * from "jobs" where "id" = ?',
                        (rec.link,))
                    rec = self.cur.fetchone()
                    assert rec is not None, "database integrity violation"
                    rec = JobEntry(*rec)

                else:
                    # link is not set, the result is still waited upon
                    raise ValueError("Result for this job is not available.")

            assert rec.result is not None, "no result found"
            # result is found! return it
            result_value = self.registry.from_json(rec.result, deref=True)
            return result_value

    def add_job_to_db(self, key, job):
        """Add job info to the database."""
        job_msg = self.registry.deep_encode(job)
        prov = prov_key(job_msg)

        def set_link(duplicate_id):
            self.cur.execute(
                'update "jobs" set "link" = ?, "status" = ? where "id" = ?',
                (duplicate_id, Status.DUPLICATE, key))

        with self.lock:
            self.cur.execute(
                'select * from "jobs" where "prov" = ? '
                'and (("result" is not null) or '
                '("session" = ? and "link" is null))',
                (prov, self.session))
            rec = self.cur.fetchone()
            rec = JobEntry(*rec) if rec is not None else None

            self.cur.execute(
                'update "jobs" set "prov" = ?, "version" = ?, "function" = ?, '
                '"arguments" = ?, "status" = ? where "id" = ?',
                (prov, job_msg['data']['hints'].get('version'),
                 json.dumps(job_msg['data']['function']),
                 json.dumps(job_msg['data']['arguments']),
                 Status.WAITING,
                 key))

            if not rec:
                # no duplicate found, go on
                return 'initialized', None

            set_link(rec.id)

            if rec.result is not None and rec.status == Status.WORKFLOW:
                # the found duplicate returned a workflow
                if rec.link is not None:
                    # link is set, so result is fully realized
                    self.cur.execute(
                        'select * from "jobs" where "id" = ?',
                        (rec.link,))
                    rec = self.cur.fetchone()
                    assert rec is not None, "database integrity violation"
                    rec = JobEntry(*rec)

                else:
                    # link is not set, the result is still waited upon
                    assert rec.session == self.session, \
                        "database integrity violation"
                    self.attached[rec.id].append(key)
                    return 'attached', None

            if rec.result is not None:
                # result is found! return it
                result_value = self.registry.from_json(rec.result, deref=True)
                result = ResultMessage(
                    key, 'retrieved', result_value, None)
                return 'retrieved', result

            if rec.session == self.session:
                # still waiting for result, attach
                self.attached[rec.id].append(key)
                return 'attached', None

    def job_exists(self, prov):
        """Check if a job exists in the database."""
        with self.lock:
            self.cur.execute('select * from "jobs" where "prov" = ?;', (prov,))
            rec = self.cur.fetchone()
            return rec is not None

    def store_result_in_db(self, result, always_cache=True):
        """Store a result in the database."""
        job = self[result.key]

        def extend_dependent_links():
            with self.lock:
                new_workflow_id = id(get_workflow(result.value))
                self.links[new_workflow_id].extend(
                    self.links[id(job.workflow)])
                del self.links[id(job.workflow)]

        def store_result(status):
            result_value_msg = self.registry.to_json(result.value)
            with self.lock:
                self.cur.execute(
                    'update "jobs" set "result" = ?, '
                    '"status" = ? where "id" = ?;',
                    (result_value_msg, status, result.key))

        def acquire_links():
            with self.lock:
                linked_keys = tuple(self.links[id(job.workflow)])
                del self.links[id(job.workflow)]

                # update links for jobs up in the call-stack (parent workflows)
                n_questions = ','.join('?' * len(linked_keys))
                self.cur.execute(
                    'update "jobs" set "link" = ? where "id" in ({});'
                    .format(n_questions),
                    (result.key,) + linked_keys)

                # jobs that were attached to the parent workflow(s) will not
                # receive the current result automatically, so we need to force
                # feed them to the scheduler.
                attached_keys = ()
                for k in linked_keys:
                    attached_keys += tuple(self.attached[k])
                    del self.attached[k]

            return attached_keys

        # if the returned job is not set to be stored, but a parent job
        # is, we still need to store the result.
        if 'store' not in job.hints and not always_cache:
            if job.is_root_node and id(job.workflow) in self.links:
                if is_workflow(result.value):
                    extend_dependent_links()
                    return ()
                else:
                    store_result(Status.LINKEE)
                    return acquire_links()
            else:
                return ()

        # if the return value is a workflow, store the workflow, and add
        # links to this job, to be updated when the resolved result comes in
        if is_workflow(result.value):
            store_result(Status.WORKFLOW)
            with self.lock:
                self.links[id(get_workflow(result.value))].append(result.key)
            if job.is_root_node:
                extend_dependent_links()
            return ()

        store_result(Status.STORED)
        with self.lock:
            attached_keys = tuple(self.attached[result.key])
            del self.attached[result.key]

        if job.is_root_node:
            attached_keys += acquire_links()

        return attached_keys

    def add_time_stamp(self, db_id, name):
        """Add a timestamp to the database."""
        with self.lock:
            self.cur.execute(
                'insert into "timestamps" ("job", "what")'
                'values (?, ?);', (db_id, name))



from pygraphviz import AGraph
from inspect import Parameter


def _sugar(s):
    s = s.replace("{", "{{").replace("}", "}}")
    if len(s) > 50:
        return s[:20] + " ... " + s[-20:]
    else:
        return s


def _format_arg_list(a, v):
    if len(a) == 0:
        if v:
            return "(\u2026)"
        else:
            return "()"

    s = "({0}{1})"
    for i in a[:-1]:
        s = s.format(
            _sugar(str(i))
            if i != Parameter.empty
            else "\u2014", ", {0}{1}")

    if v:
        return s.format("\u2026", "")

    return s.format(
        _sugar(str(a[-1]))
        if a[-1] != Parameter.empty
        else "\u2014", "")


def draw_workflow(filename, workflow):
    dot = AGraph(directed=True)  # (comment="Computing scheme")
    for i, n in workflow.nodes.items():
        dot.add_node(i, label="{0} \n {1}".format(
            n.foo.__name__, _format_arg_list(n.bound_args.args, None)))

    for i in workflow.links:
        for j in workflow.links[i]:
            dot.add_edge(i, j[0])
    dot.layout(prog='dot')

    dot.draw(filename)



from ..lib import (Connection, FlushQueue, EndOfQueue)
from .job_keeper import (JobKeeper)

from ..workflow import (
    is_workflow, get_workflow, insert_result,
    Workflow, is_node_ready)
import sys


class Job:
    def __init__(self, workflow, node_id):
        self.workflow = workflow
        self.node_id = node_id

    def __iter__(self):
        return iter((self.workflow, self.node_id))

    @property
    def node(self):
        return self.workflow.nodes[self.node_id]

    @property
    def name(self):
        return self.node.foo.__name__

    @property
    def hints(self):
        return self.node.hints

    @property
    def is_root_node(self):
        return self.node_id == self.workflow.root


class DynamicLink:
    def __init__(self, source, target, node):
        self.source = source
        self.target = target
        self.node = node

    def __iter__(self):
        return iter((self.source, self.target, self.node))


error_msg_1 = \
    "A job reported an unexpected error. If this error is not so " \
    "unexpected, consider capturing it with an error handler inside " \
    "the scheduled function or with an error handler in the scheduler " \
    "for a more graceful display and exit.\n" \
    "Infringing job: \n        {}\n"


class Scheduler:
    """
    Schedules jobs, recieves results, then schedules more jobs as they
    become ready to compute. This class communicates with a pool of workers
    by means of coroutines.
    """
    def __init__(self, verbose=False, error_handler=None, job_keeper=None):
        if job_keeper is None:
            self.jobs = JobKeeper()
        else:
            self.jobs = job_keeper
        self.dynamic_links = self.jobs.workflows
        self.count = 0
        self.key_map = {}
        self.verbose = verbose
        self.handle_error = error_handler

    def run(self, connection: Connection, master: Workflow):
        """Run a workflow.

        :param connection:
            A connection giving a sink to the job-queue and a source yielding
            results.
        :type connection: Connection

        :param master:
            The workflow.
        :type master: Workflow
        """
        # initiate worker slave army and take up reins ...
        source, sink = connection.setup()

        # schedule work
        self.add_workflow(master, master, master.root, sink)
        graceful_exit = False
        errors = []

        # process results
        for job_key, status, result, err_msg in source:
            wf, n = self.jobs[job_key]
            if status == 'error':
                graceful_exit = True
                errors.append(err_msg)

                try:
                    sink.send(FlushQueue)
                except StopIteration:
                    pass

                print("Uncaught error running job: {}, {}".format(n, err_msg),
                      file=sys.stderr)
                print("Flushing queue and waiting for threads to close.",
                      file=sys.stderr, flush=True)

            if status == 'aborted':
                print("Job {} got aborted: {}".format(n, err_msg),
                      file=sys.stderr)
                print("Flushing queue and waiting for threads to close.",
                      file=sys.stderr, flush=True)
                graceful_exit = True
                errors.append(err_msg)
                try:
                    sink.send(FlushQueue)
                except StopIteration:
                    pass

            if self.verbose:
                print("sched result [{0}]: ".format(self.key_map[job_key]),
                      result,
                      file=sys.stderr, flush=True)

            del self.jobs[job_key]
            if len(self.jobs) == 0 and graceful_exit:
                for error in errors:
                    print("Exception of type", type(error), ":")
                    print(error)
                raise errors[0]

            # if this result is the root of a workflow, pop to parent
            # we do this before scheduling a child workflow, as to
            # achieve tail-call elimination.
            while n == wf.root and wf is not master:
                child = id(wf)
                _, wf, n = self.dynamic_links[child]
                del self.dynamic_links[child]

            # if we retrieve a workflow, push a child
            if is_workflow(result) and not graceful_exit:
                child_wf = get_workflow(result)
                self.add_workflow(child_wf, wf, n, sink)
                continue

            # insert the result in the nodes that need it
            wf.nodes[n].result = result
            for (tgt, address) in wf.links[n]:
                insert_result(wf.nodes[tgt], address, result)
                if is_node_ready(wf.nodes[tgt]) and not graceful_exit:
                    self.schedule(Job(workflow=wf, node_id=tgt), sink)

            # see if we're done
            if wf == master and n == master.root:
                try:
                    sink.send(EndOfQueue)
                except StopIteration:
                    pass

                return result

    def schedule(self, job, sink):
        sink.send(self.jobs.register(job))

    def add_workflow(self, wf, target, node, sink):
        self.dynamic_links[id(wf)] = DynamicLink(
            source=wf, target=target, node=node)

        for n in wf.nodes:
            if is_node_ready(wf.nodes[n]):
                self.schedule(Job(workflow=wf, node_id=n), sink)



from noodles.prov.workflow import (set_global_provenance)
from noodles.tutorial import (add, sub, mul)
from noodles.serial import base


def test_global_prov():
    a = add(4, 5)
    b = sub(a, 3)
    c = mul(b, 7)
    d = mul(b, 7)
    e = sub(c, 3)
    f = sub(c, 5)

    set_global_provenance(c._workflow, base())

    assert c._workflow.prov is not None
    assert b._workflow.prov is not None
    assert c._workflow.prov != b._workflow.prov

    set_global_provenance(d._workflow, base())
    set_global_provenance(e._workflow, base())
    set_global_provenance(f._workflow, base())

    assert c._workflow.prov == d._workflow.prov
    assert b._workflow.prov != e._workflow.prov
    assert f._workflow.prov != e._workflow.prov



import hashlib

try:
    import ujson as json
except ImportError:
    import json


def update_object_hash(m, obj):
    r = json.dumps(obj, sort_keys=True)
    m.update(r.encode())
    return m


def prov_key(job_msg, extra=None):
    """Retrieves a MD5 sum from a function call. This takes into account the
    name of the function, the arguments and possibly a version number of the
    function, if that is given in the hints.
    This version can also be auto-generated by generating an MD5 hash from the
    function source. However, the source-code may not always be reachable, or
    the result may depend on an external process which has its own
    versioning."""
    m = hashlib.md5()
    update_object_hash(m, job_msg['data']['function'])
    update_object_hash(m, job_msg['data']['arguments'])

    if 'version' in job_msg['data']['hints']:
        update_object_hash(m, job_msg['data']['hints']['version'])

    if extra is not None:
        update_object_hash(m, extra)

    return m.hexdigest()



import noodles
from noodles.tutorial import add, sub, mul, accumulate
from noodles import run_single


def test_get_intermediate_results():
    r1 = add(1, 1)
    r2 = sub(3, r1)

    def foo(a, b, c):
        return mul(add(a, b), c)

    r3 = [foo(i, r2, r1) for i in range(6)]
    r4 = accumulate(noodles.gather_all(r3))

    run_single(r4)

    assert noodles.result(r1) == 2
    assert noodles.result(r2) == 1
    assert [noodles.result(r) for r in r3] == [2, 4, 6, 8, 10, 12]
    assert noodles.result(r4) == 42



import time

from noodles import schedule, schedule_hint, gather
from noodles.workflow import get_workflow
from noodles.run.scheduler import Scheduler
from noodles.run.worker import run_job
from noodles.run.hybrid import hybrid_coroutine_worker
from noodles.run.hybrid import run_hybrid

from noodles.lib import (Queue, Connection, pull, thread_pool)
from noodles.run.worker import (worker)
from itertools import repeat


def single_worker():
    return Queue() >> worker


def threaded_worker(n_threads):
    return Queue() >> thread_pool(*repeat(worker, n_threads))


@schedule_hint(n=1)
def f(x):
    return 2*x


@schedule_hint(n=2)
def g(x):
    return 3*x


@schedule
def h(x, y):
    return x + y


def selector(job):
    if job.hints and 'n' in job.hints:
        return job.hints['n']
    else:
        return None


@schedule_hint(n=1)
def delayed(a, dt):
    time.sleep(dt)
    return a


@schedule
def sum(a, buildin_sum=sum):
    return buildin_sum(a)


def test_hybrid_threaded_runner_02():
    A = [delayed(1, 0.01) for i in range(8)]
    B = sum(gather(*A))

    start = time.time()
    assert run_hybrid(B, selector, {1: threaded_worker(8)}) == 8
    end = time.time()
    assert (end - start) < 0.05


def test_hybrid_threaded_runner_03():
    A = [delayed(1, 0.01) for i in range(8)]
    B = sum(gather(*A))

    start = time.time()
    assert run_hybrid(B, selector, {1: single_worker()}) == 8
    end = time.time()
    assert (end - start) > 0.08


def tic_worker(tic):
    jobs = Queue()

    @pull
    def get_result():
        source = jobs.source()

        for key, job in source:
            tic()
            yield run_job(key, job)

    return Connection(get_result, jobs.sink)


def ticcer():
    a = 0

    def f():
        nonlocal a
        a += 1
        return a

    def g():
        nonlocal a
        return a

    return f, g


def test_hybrid_coroutine_runner_03():
    A1 = [f(i) for i in range(11)]
    A2 = [g(i) for i in range(7)]
    B = sum(gather(*(A1+A2)))

    tic1, c1 = ticcer()
    tic2, c2 = ticcer()

    result = Scheduler().run(
        hybrid_coroutine_worker(selector, {
            1: tic_worker(tic1), 2: tic_worker(tic2)}),
        get_workflow(B))

    assert c1() == 11
    assert c2() == 7
    assert result == 173



import queue
from .streams import (push, pull)
from .connection import Connection


class EndOfQueue(object):
    pass


class FlushQueue(object):
    pass


class Queue(Connection):
    """A |Queue| object hides a :py:class:`queue.Queue` object
    behind a source and sink interface.

    .. py:attribute:: sink

        Receives items that are put on the queue. Pushing the `end-of-queue`
        message through the sink will put it on the queue, and will also result
        in a :py:exc:`StopIteration` exception being raised.

    .. py:attribute:: source

        Pull items from the queue. When `end-of-queue` is encountered the
        generator returns after re-inserting the `end-of-queue` message on the
        queue for other sources to pick up. This way, if many threads are
        pulling from this queue, they all get the `end-of-queue` message.

    .. |Queue| replace:: :py:class:`Queue`
    """
    def __init__(self, end_of_queue=EndOfQueue):
        """
        :param end_of_queue: When this object is encountered in both
        the sink and the source, their respective loops are terminated.
        Equality is checked using `is`; usualy this is a stub class
        designed especially for this purpose.
        """
        self._queue = queue.Queue()
        self._end_of_queue = end_of_queue
        self._flush_queue = FlushQueue

        @push
        def sink():
            while True:
                r = yield
                self._queue.put(r)

                if r is self._flush_queue:
                    self.flush()
                    return

                if r is self._end_of_queue:
                    return

        @pull
        def source():
            while True:
                v = self._queue.get()
                if v is self._end_of_queue:
                    self._queue.task_done()
                    self._queue.put(self._end_of_queue)
                    return

                yield v
                self._queue.task_done()

        super(Queue, self).__init__(source, sink)

    def flush(self):
        """Erases queue and set `end-of-queue` message."""
        while not self._queue.empty():
            self._queue.get()
            self._queue.task_done()
        self.close()

    def close(self):
        """Sends `end_of_queue` message to the queue.
        Doesn't stop running sinks."""
        self._queue.put(self._end_of_queue)

    def empty(self):
        return self._queue.empty()

    def wait(self):
        self._queue.join()



"""
Test available runners on a variety of workflows.
"""

import pytest


def test_run(workflow, backend):
    if not backend.supports(workflow.requires):
        pytest.skip("Workflow not supported on this backend.")

    if workflow.raises is not None:
        with pytest.raises(workflow.raises):
            backend.run(workflow.make())
    else:
        result = backend.run(workflow.make())
        workflow.check_assertions(result)



from .pretty_term import OutStream
import sys


class Display:
    """A modest display to track jobs being run. The message being printed
    for each job depends on the hints present. This display supports two
    keywords: 'display' and 'check'.

    'display' Gives a string that will be formatted with the arguments to
    the function call.

    'check' Makes sure that the user gets feedback on the success or failure
    of the job, meaning a green V or red X will appear behind the ."""
    def __init__(self, error_filter):
        self.jobs = {}
        self.out = OutStream(sys.stdout)
        self.errors = []
        self.error_filter = error_filter
        self.messages = []

    def start(self, key, job, _):
        if job.hints and 'display' in job.hints:
            msg = job.hints['display'].format(**job.bound_args.arguments)[:70]
            self.add_job(key, job, msg)
            self.out << msg << "\n"

    def done(self, key, data, msg):
        if key in self.jobs and 'confirm' in self.jobs[key]:
            self.out << ['save'] << ['up', self.jobs[key]['line']] \
                << ['forward', max(50, self.jobs[key]['length'] + 2)]
            self.out << "(" << ['fg', 60, 180, 100] << "✔" << ['reset'] \
                << ")" << ['restore']

        if key in self.jobs and msg:
            self.message_handler(self.jobs[key], msg)

    def error(self, key, _, data):
        if key in self.jobs and 'confirm' in self.jobs[key]:
            self.out << ['save'] << ['up', self.jobs[key]['line']] \
                << ['forward', max(50, self.jobs[key]['length'] + 2)]
            self.out << "(" << ['fg', 240, 100, 60] << "✘" << ['reset'] \
                << ")" << ['restore']

    def add_job(self, key, job, msg):
        for k in self.jobs:
            self.jobs[k]['line'] += 1
        self.jobs[key] = {'line': 1, 'job': job, 'length': len(msg)}
        self.jobs[key].update(job.hints)

    def error_handler(self, job, xcptn):
        self.errors.append((job, xcptn))

    def message_handler(self, job, warning):
        self.messages.append((job['job'], warning))

    def report(self):
        if len(self.errors) == 0:
            self.out << "╰─(success)\n"

            if len(self.messages) != 0:
                self.out << "There were warnings: \n\n"

                for job, w in self.messages:
                    msg = 'WARNING '
                    if job.hints and 'display' in job.hints:
                        msg += job.hints['display'].format(
                            **job.bound_args.arguments)
                    else:
                        msg += 'calling {} with {}'.format(
                            job.foo.__name__, dict(job.bound_args.arguments)
                        )
                    print(msg)
                    print(w)

        else:
            self.out << "╰─(" << ['fg', 240, 100, 60] << "ERROR!" \
                     << ['reset'] << ")\n\n"

            for job, e in self.errors:
                msg = 'ERROR '
                if job.hints and 'display' in job.hints:
                    msg += job.hints['display'].format(
                        **job.bound_args.arguments)
                else:
                    msg += 'calling {} with {}'.format(
                        job.foo.__name__, dict(job.bound_args.arguments)
                    )

                print(msg)
                err_msg = self.error_filter(e)
                if err_msg:
                    print(err_msg)
                else:
                    print(e)

    def __call__(self, q):
        self.q = q
        for status, key, data, err_msg in q.source():
            getattr(self, status)(key, data, err_msg)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.wait()

        if exc_type:
            if exc_type is KeyboardInterrupt:
                self.out << "\n" << ['fg', 255, 200, 50] \
                         << "User interrupt detected, abnormal exit.\n" \
                         << ['reset']
                return True

            print("Internal error encountered. Contact the developers.")
            return False

        self.report()

    def wait(self):
        self.q.wait()



from ..worker import (worker)
from ..scheduler import (Scheduler)
from ...lib import (Queue)
from ...workflow import (get_workflow)


def run_single(workflow):
    """"Run workflow in a single thread (same as the scheduler).

    :param workflow: Workflow or PromisedObject to be evaluated.
    :return: Evaluated result.
    """
    return Scheduler().run(
        Queue() >> worker,
        get_workflow(workflow))



from ...lib import (
    Queue, Connection, EndOfQueue, push, pull, pull_map, sink_map,
    thread_counter)

from ..messages import (
    JobMessage, ResultMessage, EndOfWork)

import threading
import sys
from collections import namedtuple

from .xenon import (Machine)
import grpc


def xenon_interactive_worker(
        machine, worker_config, input_queue=None, stderr_sink=None):
    """Uses Xenon to run a single remote interactive worker.

    Jobs are read from stdin, and results written to stdout.

    :param machine:
        Specification of the machine on which to run.
    :type machine: noodles.run.xenon.Machine

    :param worker_config:
        Job configuration. Specifies the command to be run remotely.
    :type worker_config: noodles.run.xenon.XenonJobConfig
    """
    if input_queue is None:
        input_queue = Queue()

    registry = worker_config.registry()

    @pull_map
    def serialise(obj):
        """Serialise incoming objects, yielding strings."""
        if isinstance(obj, JobMessage):
            print('serializing:', str(obj.node), file=sys.stderr)
        return (registry.to_json(obj, host='scheduler') + '\n').encode()

    @pull_map
    def echo(line):
        print('{} input: {}'.format(worker_config.name, line), file=sys.stderr)
        return line

    def do_iterate(source):
        for x in source():
            if x is EndOfQueue:
                yield EndOfWork
                return
            yield x

    job, output_stream = machine.scheduler.submit_interactive_job(
        worker_config.xenon_job_description,
        echo(lambda: serialise(lambda: do_iterate(input_queue.source))))

    @sink_map
    def echo_stderr(text):
        """Print lines."""
        for line in text.split('\n'):
            print("{}: {}".format(worker_config.name, line), file=sys.stderr)

    if stderr_sink is None:
        stderr_sink = echo_stderr()

    @pull
    def read_output(source):
        """Handle output from job, sending stderr data to given
        `stderr_sink`, passing on lines from stdout."""
        line_buffer = ""
        try:
            for chunk in source():
                if chunk.stdout:
                    lines = chunk.stdout.decode().splitlines(keepends=True)

                    if not lines:
                        continue

                    if lines[0][-1] == '\n':
                        yield line_buffer + lines[0]
                        line_buffer = ""
                    else:
                        line_buffer += lines[0]

                    if len(lines) == 1:
                        continue

                    yield from lines[1:-1]

                    if lines[-1][-1] == '\n':
                        yield lines[-1]
                    else:
                        line_buffer = lines[-1]

                if chunk.stderr:
                    for line in chunk.stderr.decode().split('\n'):
                        stripped_line = line.strip()
                        if stripped_line != '':
                            stderr_sink.send(stripped_line)

        except grpc.RpcError as e:
            return

    @pull_map
    def deserialise(line):
        result = registry.from_json(line, deref=False)
        return result

    return Connection(
        lambda: deserialise(lambda: read_output(lambda: output_stream)),
        input_queue.sink, aux=job)


class XenonInteractiveWorker(Connection):
    """Submits a new job to the specified queue. Starts a thread to
    read and forward stderr. Then acts as a `Connection` that goes
    online as soon as the submitted job is running.
    """
    def __init__(self, machine, worker_config):
        connection = xenon_interactive_worker(machine, worker_config)
        super(XenonInteractiveWorker, self).__init__(
                connection.source, connection.sink)

        self.worker_config = worker_config
        self.job = connection.aux
        self.machine = machine
        self.online = False

    def wait_until_running(self, callback=None):
        """Waits until the remote worker is running, then calls the callback.
        Usually, this method is passed to a different thread; the callback
        is then a function patching results through to the result queue."""
        status = self.machine.scheduler.wait_until_running(
                self.job, self.worker_config.time_out)

        if status.running:
            self.online = True
            if callback:
                callback(self)
        else:
            raise TimeoutError("Timeout while waiting for worker to run: " +
                               self.worker_config.name)

    def close(self):
        try:
            self.sink.send(EndOfQueue)
        except StopIteration:
            pass


RemoteWorker = namedtuple('RemoteWorker', [
    'name', 'lock', 'max', 'jobs', 'source', 'sink'])


class DynamicPool(Connection):
    """Keeps a pool of Xenon workers. At each instant, when a job is sent to
    the recieving coroutine associated with this `Connection`, we look if
    there is a worker idle and send it there. If no worker is idle, the job
    is pushed on the queue. Whenever a worker returns a result, we replennish
    its list of jobs from the queue. In this way we can supply a variable
    amount of workers, each with their own preferred amount of jobs. As of
    yet, we have no mechanism to distribute the jobs between workers in any
    way more knowledgable than that."""
    def __init__(self, machine: Machine):
        self.workers = {}
        self.machine = machine

        self.job_queue = Queue()
        self.result_queue = Queue()
        self.wlock = threading.Lock()
        self.plock = threading.Lock()
        self.count = thread_counter(self.result_queue.close)

        @push
        def dispatch_jobs():
            job_sink = self.job_queue.sink()

            while True:
                msg = yield

                for w in self.workers.values():
                    with w.lock:  # Worker lock ~~~~~~~~~~~~~~~~~~~~~
                        if len(w.jobs) < w.max:
                            w.sink.send(msg)
                            w.jobs.append(msg.key)
                            break
                    # lock end ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                else:
                    job_sink.send(msg)

        super(DynamicPool, self).__init__(
            self.result_queue.source, dispatch_jobs)

    def close_all(self):
        for w in self.workers.values():
            try:
                w.sink.send(EndOfQueue)
            except StopIteration:
                pass

    def add_xenon_worker(self, worker_config):
        """Adds a worker to the pool; sets gears in motion."""
        c = XenonInteractiveWorker(self.machine, worker_config)
        w = RemoteWorker(
            worker_config.name, threading.Lock(),
            worker_config.n_threads, [], *c.setup())

        with self.wlock:
            self.workers[worker_config.name] = w

        def populate(job_source):
            """Populate the worker with jobs, if jobs are available."""
            with w.lock, self.plock:  # Worker lock ~~~~~~~~~~~~~~~~~~
                while len(w.jobs) < w.max and not self.job_queue.empty():
                    msg = next(job_source)
                    w.sink.send(msg)

                    if msg is not EndOfQueue:
                        key, job = msg
                        w.jobs.append(key)
                    else:
                        break
            # lock end ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        def activate(_):
            """Activate the worker."""
            job_source = self.job_queue.source()
            populate(job_source)

            sink = self.result_queue.sink()

            for result in w.source:
                sink.send(result)

                # do bookkeeping and submit a new job to the worker
                with w.lock:  # Worker lock ~~~~~~~~~~~~~~~~~~~~~
                    w.jobs.remove(result.key)
                populate(job_source)
                # lock end ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

            for key in w.jobs:
                sink.send(ResultMessage(
                    key, 'aborted', None, 'connection to remote worker lost.'))

        # Start the `activate` function when the worker goes online.
        t = threading.Thread(
            target=self.count(c.wait_until_running), args=(activate,),
            daemon=True)
        t.start()



"""
(stub)

Noodles
=======
"""

from noodles.interface import (
    delay, gather, lift, schedule, schedule_hint, unwrap,
    has_scheduled_methods, update_hints, unpack, quote, unquote,
    gather_dict, result, gather_all, maybe, Fail, simple_lift, ref, failed)

from noodles.workflow import (get_workflow)
from .patterns import (fold, find_first, conditional)
from .run.runners import run_parallel_with_display as run_logging

from .run.threading.vanilla import run_parallel
from .run.single.vanilla import run_single

from .run.process import run_process
from .run.scheduler import Scheduler

__version__ = "0.3.0"

__all__ = ['schedule', 'schedule_hint', 'run_single', 'run_process',
           'Scheduler', 'has_scheduled_methods',
           'Fail', 'failed',
           'run_logging', 'run_parallel', 'unwrap', 'get_workflow',
           'gather', 'gather_all', 'gather_dict', 'lift', 'unpack',
           'maybe', 'delay', 'update_hints',
           'quote', 'unquote', 'result', 'fold', 'find_first',
           'conditional',
           'simple_lift', 'ref']



from functools import wraps


def coroutine(f):
    """
    A sink should be send `None` first, so that the coroutine arrives
    at the `yield` position. This wrapper takes care that this is done
    automatically when the coroutine is started.
    """
    @wraps(f)
    def g(*args, **kwargs):
        sink = f(*args, **kwargs)
        sink.send(None)
        return sink

    return g



from .pretty_term import OutStream
import sys


class Display:
    """A modest display to track jobs being run. The message being printed
    for each job depends on the hints present. This display supports two
    keywords: 'display' and 'check'.

    'display' Gives a string that will be formatted with the arguments to
    the function call.

    'check' Makes sure that the user gets feedback on the success or failure
    of the job, meaning a green V or red X will appear behind the ."""
    def __init__(self, error_filter=None, title='running jobs'):
        self.jobs = {}
        self.out = OutStream(sys.stdout)
        self.errors = []
        self.error_filter = error_filter or (lambda e: None)
        self.messages = []

        ascii_chars = {
            'check':        'V',
            'fail':         'X',
            'line-hor':     '-',
            'line-ver':     '|',
            'line-tl':      '+',
            'line-bl':      '+'
        }

        utf8_chars = {
            'check':        '✔',
            'fail':         '✘',
            'line-hor':     '─',
            'line-ver':     '│',
            'line-tl':      '╭',
            'line-bl':      '╰'
        }

        self.chars = utf8_chars \
            if sys.stdout.encoding == 'UTF-8' \
            else ascii_chars

        self.out << self.chars['line-tl'] << self.chars['line-hor'] \
                 << '(' << title << ')' << "\n"

    def start(self, key, job, _):
        if job.hints and 'display' in job.hints:
            if key not in self.jobs:
                msg = job.hints['display'].format(
                    **job.bound_args.arguments)[:70]
                self.add_job(key, job, msg)
                self.out << self.chars['line-ver'] << "    " << msg << "\n"
            else:
                self.out << ['save'] << ['up', self.jobs[key]['line']] \
                    << ['forward', max(50, self.jobs[key]['length'] + 2)]
                self.out << "( )" << ['restore']

    def done(self, key, data, msg):
        if key in self.jobs and 'confirm' in self.jobs[key]:
            self.out << ['save'] << ['up', self.jobs[key]['line']] \
                << ['forward', max(50, self.jobs[key]['length'] + 2)]
            self.out << "(" << ['fg', 60, 180, 100] << self.chars['check'] \
                << ['reset'] << ")" << ['restore']

        if key in self.jobs and msg:
            self.message_handler(self.jobs[key], msg)

    def schedule(self, key, job, _):
        if job.hints and 'display' in job.hints:
            msg = job.hints['display'].format(**job.bound_args.arguments)[:70]
            self.add_job(key, job, msg)
            self.out << self.chars['line-ver'] << " " << msg << "\n"

    def retrieved(self, key, data, msg):
        if key in self.jobs and 'confirm' in self.jobs[key]:
            self.out << ['save'] << ['up', self.jobs[key]['line']] \
                << ['forward', max(50, self.jobs[key]['length'] + 2)]
            self.out << "(" << ['fg', 60, 180, 100] << "retrieved" \
                << ['reset'] << ")" << ['restore']

        if key in self.jobs and msg:
            self.message_handler(self.jobs[key], msg)

    def error(self, key, _, data):
        if key in self.jobs and 'confirm' in self.jobs[key]:
            self.out << ['save'] << ['up', self.jobs[key]['line']] \
                << ['forward', max(50, self.jobs[key]['length'] + 2)]
            self.out << "(" << ['fg', 240, 100, 60] << self.chars['fail'] \
                << ['reset'] << ")" << ['restore']

    def add_job(self, key, job, msg):
        for k in self.jobs:
            self.jobs[k]['line'] += 1
        self.jobs[key] = {'line': 1, 'job': job, 'length': len(msg)}
        self.jobs[key].update(job.hints)

    def error_handler(self, job, *exception):
        msg = self.error_filter(*exception)
        if msg:
            self.errors.append((job, msg))
            return True
        else:
            return False

    def message_handler(self, job, warning):
        self.messages.append((job['job'], warning))

    def report(self):
        if len(self.errors) == 0:
            self.out << self.chars['line-bl'] << self.chars['line-hor'] \
                     << "(success)\n"

            if len(self.messages) != 0:
                self.out << "There were warnings: \n\n"

                for job, w in self.messages:
                    msg = 'WARNING '
                    if job.hints and 'display' in job.hints:
                        msg += job.hints['display'].format(
                            **job.bound_args.arguments)
                    else:
                        msg += 'calling {} with {}'.format(
                            job.foo.__name__, dict(job.bound_args.arguments)
                        )
                    print(msg)
                    print(w)

        else:
            self.out << self.chars['line-bl'] << self.chars['line-hor'] \
                     << "(" << ['fg', 240, 100, 60] << "ERROR!" \
                     << ['reset'] << ")\n\n"

            for job, e in self.errors:
                msg = 'ERROR '
                if job.hints and 'display' in job.hints:
                    msg += job.hints['display'].format(
                        **job.bound_args.arguments)
                else:
                    msg += 'calling {} with {}'.format(
                        job.foo.__name__, dict(job.bound_args.arguments)
                    )

                print(msg)
                print(e)

    def __call__(self, msg):
        key, status, data, err_msg = msg
        getattr(self, status)(key, data, err_msg)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # self.wait()

        if exc_type:
            if exc_type is KeyboardInterrupt:
                self.out << "\n" << ['fg', 255, 200, 50] \
                         << "User interrupt detected, abnormal exit.\n" \
                         << ['reset']
                return True

            if exc_type is SystemExit:
                return False

            print("Internal error encountered. Contact the developers: \n",
                  exc_type, exc_val)
            return False

        self.report()



# import pytest

from noodles.prov.sqlite import JobDB
from noodles import serial
from noodles.tutorial import (sub, add)
from noodles.run.scheduler import Job
from noodles.run.messages import (ResultMessage)


def test_add_job():
    db = JobDB(':memory:', registry=serial.base)

    wf = sub(1, 1)
    job = Job(wf._workflow, wf._workflow.root)
    key, node = db.register(job)
    msg, value = db.add_job_to_db(key, node)
    assert msg == 'initialized'

    duplicates = db.store_result_in_db(ResultMessage(key, 'done', 0, None))
    assert duplicates == ()

    key, node = db.register(job)
    msg, result = db.add_job_to_db(key, node)
    assert msg == 'retrieved'
    assert result.value == 0


def test_attaching():
    db = JobDB(':memory:', registry=serial.base)

    wf = add(1, 1)
    job = Job(wf._workflow, wf._workflow.root)
    key1, node1 = db.register(job)
    msg, value = db.add_job_to_db(key1, node1)
    assert msg == 'initialized'

    key2, node2 = db.register(job)
    msg, value = db.add_job_to_db(key2, node2)
    assert msg == 'attached'

    duplicates = db.store_result_in_db(ResultMessage(key1, 'done', 2, None))
    assert duplicates == (key2,)

    key3, node3 = db.register(job)
    msg, result = db.add_job_to_db(key3, node3)
    assert msg == 'retrieved'
    assert result.value == 2



"""
Functions useful for tutorial work and unit testing.
"""

from inspect import Parameter
from graphviz import Digraph
import html
import sys

from . import schedule, schedule_hint, get_workflow


@schedule
def echo_add(x, y):
    print('adding', x, 'and', y, file=sys.stderr)
    return x + y


@schedule
def add(x, y):
    """add `x` and `y`."""
    return x + y


@schedule_hint(display="{a} + {b}", confirm=True)
def log_add(a, b):
    """add `a` and `b` and send message to logger."""
    return a + b


@schedule
def sub(x, y):
    """subtract `y` from `x`."""
    return x - y


@schedule
def mul(x, y):
    """multiply `x` and `y`."""
    return x*y


@schedule
def accumulate(lst, start=0):
    """compute sum of `lst`."""
    return sum(lst, start)


def _format_arg_list(args, variadic=False):
    """Format a list of arguments for pretty printing.

    :param a: list of arguments.
    :type a: list

    :param v: tell if the function accepts variadic arguments
    :type v: bool
    """
    def sugar(s):
        """Shorten strings that are too long for decency."""
        s = s.replace("{", "{{").replace("}", "}}")
        if len(s) > 50:
            return s[:20] + " ... " + s[-20:]
        else:
            return s

    def arg_to_str(arg):
        """Convert argument to a string."""
        if isinstance(arg, str):
            return sugar(repr(arg))
        elif arg is Parameter.empty:
            return '\u2014'
        else:
            return sugar(str(arg))

    if not args:
        if variadic:
            return "(\u2026)"
        else:
            return "()"

    return "(" + ", ".join(map(arg_to_str, args)) + ")"


def get_workflow_graph(promise):
    """Get a graph of a promise."""
    workflow = get_workflow(promise)

    dot = Digraph()
    for i, n in workflow.nodes.items():
        dot.node(str(i), label="{0} \n {1}".format(
            n.foo.__name__,
            _format_arg_list(n.bound_args.args)))

    for i in workflow.links:
        for j in workflow.links[i]:
            dot.edge(str(i), str(j[0]), label=str(j[1].name))

    return dot


def display_workflows(prefix, **kwargs):
    """Display workflows in a table. This generates SVG files. Use this in a
    Jupyter notebook and ship the images

    :param prefix: name prefix for svg files generated.
    :param kwargs: keyword arguments containing a workflow each.
    """
    from IPython.display import display, Markdown

    def create_svg(name, workflow):
        """Create an SVG file with rendered graph from a workflow."""
        name = name.replace('_', '-')
        filename = '{}-{}.svg'.format(prefix, name)
        dot = get_workflow_graph(workflow)
        dot.attr('graph', bgcolor='transparent')
        with open(filename, 'bw') as file:
            file.write(dot.pipe(format='svg'))
        return name, filename

    svg_files = {name: svg for name, svg in
                 (create_svg(name, workflow)
                  for name, workflow in kwargs.items())}

    markdown_table = \
        '| ' + ' | '.join(svg_files.keys()) + ' |\n' + \
        '| ' + ' | '.join('---' for _ in svg_files) + ' |\n' + \
        '| ' + ' | '.join('![workflow {}]({})'.format(name, svg)
                          for name, svg in svg_files.items()) + ' |'
    display(Markdown(markdown_table))


def snip_line(line, max_width, split_at):
    """Shorten a line to a maximum length."""
    if len(line) < max_width:
        return line
    return line[:split_at] + " … " \
        + line[-(max_width - split_at - 3):]


def display_text(text, highlight=None, max_width=80, split_at=None):
    """In a Jupyter notebook, takes a multi-line string, displays these lines
    in '<pre></pre>' block HTML. A list of lines may be provided that should
    be highlighted in the result. A highlight gives a gray background and bold
    font.

    :param text: string
    :param lines: list of lines to be highlighted.
    """
    from IPython.display import display, HTML

    if not isinstance(text, str):
        text = str(text)

    highlight = highlight or []
    split_at = split_at or max_width - 10
    style = [
        'font-size: 9pt; margin: 0pt',
        'background: #eeeeee; color: black;'
        'font-weight: bold; font-size: 9pt; margin: 0pt']
    display(HTML('\n'.join(
        '<pre style="{}">{}</pre>'.format(
            style[i in highlight],
            html.escape(snip_line(line, max_width, split_at)))
        for i, line in enumerate(text.splitlines()))))


def run_and_print_log(workflow, highlight=None):
    """Run workflow on multi-threaded worker cached with Sqlite3.

    :param workflow: workflow to evaluate.
    :param highlight: highlight these lines.
    """
    from noodles.run.threading.sqlite3 import run_parallel
    from noodles import serial

    import io
    import logging

    log = io.StringIO()
    log_handler = logging.StreamHandler(log)
    formatter = logging.Formatter('%(asctime)s - %(message)s')
    log_handler.setFormatter(formatter)

    logger = logging.getLogger('noodles')
    logger.setLevel(logging.INFO)
    logger.handlers = [log_handler]

    result = run_parallel(
        workflow, n_threads=4, registry=serial.base, db_file='tutorial.db',
        always_cache=True, echo_log=False)
    display_text(log.getvalue(), highlight or [], split_at=40)
    return result



"""
Maybe
=====

Facility to handle non-fatal errors in Noodles.
"""

from functools import (wraps)
from itertools import (chain)
import inspect
from ..lib import (object_name)


class Fail:
    """Signifies a failure in a computation that was wrapped by a ``@maybe``
    decorator. Because Noodles runs all functions from the same context, it
    is not possible to use Python stack traces to find out where an error
    happened. In stead we use a ``Fail`` object to store information about
    exceptions and the subsequent continuation of the failure."""
    def __init__(self, func, fails=None, exception=None):
        try:
            self.name = "{} ({}:{})".format(
                object_name(func),
                inspect.getsourcefile(func),
                inspect.getsourcelines(func)[1])
        except AttributeError:
            self.name = "<{} instance>".format(func.__class__.__name__)

        self.fails = fails or []
        self.trace = []
        self.exception = exception

    def add_call(self, func):
        """Add a call to the trace."""
        self.trace.append("{} ({}:{})".format(
            object_name(func),
            inspect.getsourcefile(func),
            inspect.getsourcelines(func)[1]))

        return self

    @property
    def is_root_cause(self):
        """If the field ``exception`` is set in this object, it means
        that we are looking at the root cause of the failure."""
        return self.exception is not None

    def __bool__(self):
        return False

    def __str__(self):
        msg = "Fail: " + " -> ".join(self.trace + [self.name])
        if self.exception is not None:
            msg += "\n* {}: ".format(type(self.exception).__name__)
            msg += "\n    ".join(l for l in str(self.exception).split('\n'))
        elif self.fails:
            msg += "\n* failed arguments:\n    "
            msg += "\n    ".join(
                "{} `{}` ".format(func, source) + "\n    ".join(
                    l for l in str(fail).split('\n'))
                for func, source, fail in self.fails)
        return msg


def failed(obj):
    """Returns True if ``obj`` is an instance of ``Fail``."""
    return isinstance(obj, Fail)


def maybe(func):
    """Calls `f` in a try/except block, returning a `Fail` object if
    the call fails in any way. If any of the arguments to the call are Fail
    objects, the call is not attempted."""

    name = object_name(func)

    @wraps(func)
    def maybe_wrapped(*args, **kwargs):
        """@maybe wrapped version of ``func``."""
        fails = [
            (name, k, v)
            for k, v in chain(enumerate(args), kwargs.items())
            if isinstance(v, Fail)]

        if fails:
            return Fail(func, fails=fails)

        try:
            result = func(*args, **kwargs)

        except Exception as exc:
            return Fail(func, exception=exc)

        else:
            if isinstance(result, Fail):
                result.add_call(func)

            return result

    return maybe_wrapped



from .control_flow_statements import (
    conditional)
from .find_first import (find_first)
from .functional_patterns import (all, any, filter, fold, map, zip_with)

__all__ = [
    'all', 'any', 'filter', 'fold', 'find_first', 'conditional', 'map',
    'zip_with']



"""
This is the Noodles executable worker module. This is usually run by the
scheduler to dispatch jobs remotely, using

.. code-block:: bash

    > python3 -m noodles.pilot_job <args>...

Recieve worker commands as JSON objects through stdin and send out results to
stdout.

    .. code-block:: bash

        > python3 -m noodles.worker -online [-use <worker>]
"""

import argparse
import sys
import uuid
from contextlib import redirect_stdout

import time
import os

from noodles import __version__
from .lib import (look_up)
from .run.messages import (EndOfWork)

from .run.worker import (
    run_job)

from .run.messages import (
    JobMessage)

from .run.remote.io import (JSONObjectReader, JSONObjectWriter)


def run_online_mode(args):
    """Run jobs.

    :param args: arguments resulting from program ArgumentParser.
    :return: None

    This reads messages containing job descriptions from standard input,
    and writes messages to standard output containing the result.

    Messages can be encoded as either JSON or MessagePack.
    """
    print("\033[47;30m Netherlands\033[48;2;0;174;239;37m▌"
          "\033[38;2;255;255;255me\u20d2Science\u20d2\033[37m▐"
          "\033[47;30mcenter \033[m Noodles worker", file=sys.stderr)

    if args.n == 1:
        registry = look_up(args.registry)()
        finish = None

        input_stream = JSONObjectReader(
            registry, sys.stdin, deref=True)
        output_stream = JSONObjectWriter(
            registry, sys.stdout, host=args.name)
        sys.stdout.flush()

        # run the init function if it is given
        if args.init:
            with redirect_stdout(sys.stderr):
                look_up(args.init)()

        if args.finish:
            finish = look_up(args.finish)

        for msg in input_stream:
            if isinstance(msg, JobMessage):
                key, job = msg
            elif msg is EndOfWork:
                print("received EndOfWork, bye", file=sys.stderr)
                sys.exit(0)
            elif isinstance(msg, tuple):
                key, job = msg
            elif msg is None:
                continue
            else:
                raise RuntimeError("Unknown message received: {}".format(msg))

            if args.jobdirs:
                # make a directory
                os.mkdir("noodles-{0}".format(key.hex))
                # enter it
                os.chdir("noodles-{0}".format(key.hex))

            if args.verbose:
                print("worker: ",
                      job.foo.__name__,
                      job.bound_args.args,
                      job.bound_args.kwargs,
                      file=sys.stderr, flush=True)

            with redirect_stdout(sys.stderr):
                result = run_job(key, job)

            if args.verbose:
                print("result: ", result.value, file=sys.stderr, flush=True)

            if args.jobdirs:
                # parent directory
                os.chdir("..")

            output_stream.send(result)

        if finish:
            finish()

        time.sleep(0.1)
        sys.stdout.flush()
        sys.stderr.flush()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="{py} -m noodles.worker".format(
            py=os.path.basename(sys.executable)),

        description="This is the Noodles executable worker module. "
                    "This is usually run by the scheduler to dispatch "
                    "jobs remotely.")

    parser.add_argument(
        "-version", action="version", version="Noodles {}".format(__version__))
    parser.add_argument(
        "-registry", type=str,
        help="the serial registry")
    parser.add_argument(
        "-n", type=int,
        help="the number of threads.", default=1)
    parser.add_argument(
        "-verbose",
        help="output information to stderr for debugging",
        default=False, action='store_true')
    parser.add_argument(
        "-jobdirs",
        help="create a directory for each job to run in",
        default=False, action='store_true')
    parser.add_argument(
        "-name", type=str,
        help="worker identity",
        default="worker-" + str(uuid.uuid4()))
    parser.add_argument(
        "-init", type=str,
        help="an init function will be send before other jobs",
        default=None)
    parser.add_argument(
        "-finish", type=str,
        help="a finish function will be send before other jobs",
        default=None)

    run_online_mode(parser.parse_args())



from inspect import (signature, Parameter)
from .model import (
    Workflow, FunctionNode, get_workflow, is_workflow)
from .arguments import (
    ref_argument, serialize_arguments, set_argument, Empty)

from copy import deepcopy


def from_call(foo, args, kwargs, hints, call_by_value=True):
    """Takes a function and a set of arguments it needs to run on. Returns a newly
    constructed workflow representing the promised value from the evaluation of
    the function with said arguments.

    These arguments are stored in a BoundArguments object matching to the
    signature of the given function ``foo``. That is, bound_args was
    constructed by doing::

        inspect.signature(foo).bind(*args, **kwargs)

    The arguments stored in the ``bound_args`` object are filtered on being
    either plain, or promised. If an argument is promised, the value
    it represents is not actually available and needs to be computed by
    evaluating a workflow.

    If an argument is a promised value, the workflow representing the value
    is added to the new workflow. First all the nodes in the original workflow,
    if not already present in the new workflow from an earlier argument,
    are copied to the new workflow, and a new entry is made into the link
    dictionary. Then the links in the old workflow are also added to the
    link dictionary. Since the link dictionary points from nodes to a
    :py:class:`set` of :py:class:`ArgumentAddress` es, no links are
    duplicated.

    In the ``bound_args`` object the promised value is replaced by the
    ``Empty`` object, so that we can see which arguments still have to be
    evaluated.

    Doing this for all promised value arguments in the bound_args object,
    results in a new workflow with all the correct dependencies represented
    as links in the graph.

    :param foo:
        Function (or object) being called.
    :type foo: Callable

    :param args:
        Normal arguments to call

    :param kwargs:
        Keyword arguments to call

    :param hints:
        Hints that can be passed to the scheduler on where or how
        to schedule this job.

    :returns:
        New workflow.
    :rtype: Workflow
    """
    # create the bound_args object
    bound_args = signature(foo).bind(*args, **kwargs)
    bound_args.apply_defaults()

    # get the name of the variadic argument if there is one
    variadic = next((x.name
                     for x in bound_args.signature.parameters.values()
                     if x.kind == Parameter.VAR_POSITIONAL), None)

    # *HACK*
    # the BoundArguments class uses a tuple to store the
    # variadic arguments. Since we need to modify them,
    # we have to replace the tuple with a list. This works, for now...
    if variadic:
        if variadic not in bound_args.arguments:
            bound_args.arguments[variadic] = []
        else:
            bound_args.arguments[variadic] = \
                list(bound_args.arguments[variadic])

    # create the node and initialise hash key
    node = FunctionNode(foo, bound_args, hints)

    # setup the new workflow
    root = id(node)
    nodes = {root: node}
    links = {root: set()}

    # walk the arguments to the function call
    for address in serialize_arguments(node.bound_args):
        arg = ref_argument(node.bound_args, address)

        # the argument may still become a workflow if it
        # has the __deepcopy__ operator overloaded to return a workflow
        call_by_ref = 'call_by_ref' in hints and \
            (hints['call_by_ref'] is True or
             address.name in hints['call_by_ref'])

        if not is_workflow(arg) and call_by_value and not call_by_ref:
            arg = deepcopy(arg)

        # if still not a workflow, we have a plain value!
        if not is_workflow(arg):
            set_argument(node.bound_args, address, arg)
            continue

        # merge the argument workflow into the new workflow
        workflow = get_workflow(arg)
        set_argument(node.bound_args, address, Empty)
        for n in workflow.nodes:
            if n not in nodes:
                nodes[n] = workflow.nodes[n]
                links[n] = set()

            links[n].update(workflow.links[n])

        links[workflow.root].add((root, address))

    return Workflow(root, nodes, links)



from noodles import (schedule, quote, unquote)


def find_first(pred, lst):
    """Find the first result of a list of promises `lst` that satisfies a
    predicate `pred`.

    :param pred: a function of one argument returning `True` or `False`.
    :param lst: a list of promises or values.
    :return: a promise of a value or `None`.

    This is a wrapper around :func:`s_find_first`. The first item on the list
    is passed *as is*, forcing evalutation. The tail of the list is quoted, and
    only unquoted if the predicate fails on the result of the first promise.

    If the input list is empty, `None` is returned."""
    if lst:
        return s_find_first(pred, lst[0], [quote(l) for l in lst[1:]])
    else:
        return None


@schedule
def s_find_first(pred, first, lst):
    """Evaluate `first`; if predicate `pred` succeeds on the result of `first`,
    return the result; otherwise recur on the first element of `lst`.

    :param pred: a predicate.
    :param first: a promise.
    :param lst: a list of quoted promises.
    :return: the first element for which predicate is true."""
    if pred(first):
        return first
    elif lst:
        return s_find_first(pred, unquote(lst[0]), lst[1:])
    else:
        return None



import uuid
import time
import json
import sys

from threading import Lock
from ..lib import (coroutine, EndOfQueue)
from .messages import (JobMessage, EndOfWork)


class JobKeeper(dict):
    def __init__(self, keep=False):
        super(JobKeeper, self).__init__()
        self.keep = keep
        self.lock = Lock()
        self.workflows = {}

    def register(self, job):
        with self.lock:
            key = str(uuid.uuid4())
            job.db_id = None
            job.log = []
            job.log.append((time.time(), 'register', None, None))
            self[key] = job

        return JobMessage(key, job.node)

    def __delitem__(self, key):
        if not self.keep:
            super(JobKeeper, self).__delitem__(key)

    def store_result(self, key, status, value, err):
        if status != 'done':
            return

        if key not in self:
            print("WARNING: store_result called but job not in registry:\n"
                  "   race condition? Not doing anything.\n", file=sys.stderr)
            return

        with self.lock:
            job = self[key]
            job.node.result = value

    @coroutine
    def message(self):
        while True:
            msg = yield

            if msg is EndOfQueue:
                return
            if msg is None:
                print("Warning: `None` received where not expected.",
                      file=sys.stderr)
                return

            key, status, value, err = msg

            with self.lock:
                if key not in self:
                    continue

                job = self[key]
                job.log.append((time.time(), status, value, err))


class JobTimer(dict):
    def __init__(self, timing_file, registry=None):
        super(JobTimer, self).__init__()
        self.workflows = {}

        if isinstance(timing_file, str):
            self.fo = open(timing_file, 'w')
            self.owner = True
        else:
            self.fo = timing_file
            self.owner = False

    def register(self, job):
        key = str(uuid.uuid4())
        job.sched_time = time.time()
        self[key] = job
        return JobMessage(key, job.node)

    def __delitem__(self, key):
        pass

    # def message(self, key, status, value, err):
    @coroutine
    def message(self):
        while True:
            msg = yield
            if msg is EndOfWork:
                return
            key, status, value, err = msg
            if hasattr(self, status):
                getattr(self, status)(key, value, err)

    def start(self, key, value, err):
        self[key].start_time = time.time()

    def done(self, key, value, err):
        job = self[key]
        now = time.time()
        if job.node.hints and 'display' in job.node.hints:
            msg_obj = {
                'description': job.node.hints['display'].format(
                    **job.node.bound_args.arguments),
                'schedule_time': time.strftime(
                    '%Y-%m-%dT%H:%M:%SZ', time.gmtime(job.sched_time)),
                'start_time': time.strftime(
                    '%Y-%m-%dT%H:%M:%SZ', time.gmtime(job.start_time)),
                'done_time': time.strftime(
                    '%Y-%m-%dT%H:%M:%SZ', time.gmtime(now)),
                'run_duration': now - job.start_time}
            self.fo.write('{record},\n'.format(record=json.dumps(
                msg_obj, indent=2)))

    def __enter__(self):
        return self

    def __exit__(self, e_type, e_value, e_tb):
        if self.owner:
            self.fo.close()



from .registry import Serialiser


class Reasonable(object):
    """A Reasonable object is an object which is most reasonably serialised
    using its `__dict__` property. To deserialise the object, we first create
    an instance using the `__new__` method, then setting the `__dict__`
    property manualy. This class is empty, it is used as a tag to designate
    other objects as reasonable."""
    pass


class SerReasonableObject(Serialiser):
    def __init__(self, cls):
        super(SerReasonableObject, self).__init__(cls)

    def encode(self, obj, make_rec):
        return make_rec(obj.__dict__)

    def decode(self, cls, data):
        obj = cls.__new__(cls)
        obj.__dict__ = data
        return obj



#!/usr/bin/env python

"""
Setup script for Noodles.
"""

from pathlib import Path
from setuptools import setup, find_packages

# Get the long description from the README file
here = Path(__file__).parent.absolute()
with (here / 'README.rst').open(encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='Noodles',
    version='0.3.0',
    description='Workflow Engine',
    long_description=long_description,
    author='Johan Hidding',
    url='https://github.com/NLeSC/noodles',
    packages=find_packages(exclude=['test*']),
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Environment :: Console',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.6',
        'Topic :: System :: Distributed Computing'],

    install_requires=['graphviz', 'ujson'],
    extras_require={
        'xenon': ['pyxenon'],
        'numpy': ['numpy', 'h5py', 'filelock'],
        'develop': [
            'pytest', 'pytest', 'coverage', 'pep8', 'numpy', 'tox',
            'sphinx', 'sphinx_rtd_theme', 'nbsphinx', 'flake8'],
    },
)



from .worker import (worker)
from .job_keeper import (JobTimer)
from .scheduler import (Scheduler)

from ..workflow import (get_workflow)
from ..lib import (
    Queue, push_map, sink_map, branch, patch,
    thread_pool)

from itertools import (repeat)
import threading


@push_map
def log_job_start(job):
    return job.key, 'start', job.node, None


@push_map
def log_job_schedule(job):
    return job.key, 'schedule', job.node, None


def run_parallel_timing(wf, n, timing_file):
    LogQ = Queue()

    with JobTimer(timing_file) as J:
        S = Scheduler(job_keeper=J)
        threading.Thread(
            target=patch,
            args=(LogQ.source, J.message),
            daemon=True).start()

        W = Queue() \
            >> branch(log_job_start >> LogQ.sink) \
            >> thread_pool(*repeat(worker, n)) \
            >> branch(LogQ.sink)

        result = S.run(W, get_workflow(wf))
        LogQ.wait()
        return result


def run_single_with_display(wf, display):
    """Adds a display to the single runner. Everything still runs in a single
    thread. Every time a job is pulled by the worker, a message goes to the
    display routine; when the job is finished the result is sent to the display
    routine."""
    S = Scheduler(error_handler=display.error_handler)
    W = Queue() \
        >> branch(log_job_start.to(sink_map(display))) \
        >> worker \
        >> branch(sink_map(display))

    return S.run(W, get_workflow(wf))


def run_parallel_with_display(wf, n_threads, display):
    """Adds a display to the parallel runner. Because messages come in
    asynchronously now, we start an extra thread just for the display
    routine."""
    LogQ = Queue()

    S = Scheduler(error_handler=display.error_handler)

    threading.Thread(
        target=patch,
        args=(LogQ.source, sink_map(display)),
        daemon=True).start()

    W = Queue() \
        >> branch(log_job_start >> LogQ.sink) \
        >> thread_pool(*repeat(worker, n_threads)) \
        >> branch(LogQ.sink)

    result = S.run(W, get_workflow(wf))
    LogQ.wait()

    return result



"""
.. default-domain:: py
"""

from .decorator import (schedule, PromisedObject)
# from .maybe import (maybe)
# from copy import deepcopy
# from ..serial.reasonable import Reasonable


class Ref:
    """The Ref object circumvents deep-copying effecting call by reference."""
    def __init__(self, obj):
        self._obj = obj

    def unref(self):
        return self._obj

    def __deepcopy__(self, memo):
        return self

    def __serialize__(self, pack):
        return pack(self._obj)

    @classmethod
    def __construct__(cls, data):
        return Ref(data)


def ref(value):
    return Ref(value)


@schedule
def delay(value):
    """Creates a promise of a given value. TODO: this function should have a
    different name."""
    return value


@schedule
def gather(*a):
    """Converts a list of promises (i.e. :class:`PromisedObject`) to a promised
    list of values."""
    return list(a)


def gather_all(a):
    """Converts an iterator of promises into a promise of a list."""
    return gather(*a)


@schedule
def gather_dict(**kwargs):
    """Creates a promise of a dictionary."""
    return dict(**kwargs)


def unpack(t, n):
    """Iterates over a promised sequence, the sequence should support random
    access by :meth:`object.__getitem__`. Also the length of the sequence
    should be known beforehand.

    :param t: a sequence.
    :param n: the length of the sequence.
    :return: an unpackable generator for the elements in the sequence."""
    return (t[i] for i in range(n))


@schedule
def set_dict(obj, d):
    """Set the :attr:`object.__dict__` of `obj`.

    :param obj: input object.
    :param d: dictionary.
    :return: the modified object."""
    if d:
        obj.__dict__ = d
    return obj


@schedule
def create_object(cls, members):
    """Promise an object of class `cls` with content `members`."""
    obj = cls.__new__(cls)
    obj.__dict__ = members
    return obj


@schedule
def make_tuple(*args):
    """Promise a tuple."""
    return args


@schedule
def make_dict(*kwargs):
    """Promise a dictionary, from a list of 2-tuples."""
    return dict(kwargs)


@schedule
def make_list(*args):
    """Promise a list explicitely; function is identical to :func:`gather`."""
    return list(args)


@schedule
def make_set(*args):
    """Promise a set."""
    return set(args)


class Quote:
    """Quote objects store the contents of a workflow, allowing the workflow to
    be passed as an argument to a higher order function without its contents
    being evaluated. Don't use this object, rather use the functions
    :func:`quote` and :func:`unquote`."""
    def __init__(self, promise):
        self.workflow = promise._workflow

    @property
    def promise(self):
        return PromisedObject(self.workflow)


def quote(promise):
    """Quote a promise."""
    if isinstance(promise, PromisedObject):
        return Quote(promise)
    return promise


def unquote(quoted):
    """Unquote a quoted promise."""
    if isinstance(quoted, Quote):
        return quoted.promise
    return quoted


@schedule
def construct_object(cls, args):
    return cls(args)


@schedule
def simple_lift(obj):
    """Create a promise from a plain object."""
    return obj


def lift(obj, memo=None):
    """Make a promise out of object `obj`, where `obj` may contain promises
    internally.

    :param obj: Any object.
    :param memo: used for internal caching (similar to :func:`deepcopy`).

    If the object is a :class:`PromisedObject`, or *pass-by-value*
    (:class:`str`, :class:`int`, :class:`float`, :class:`complex`) it is
    returned as is.

    If the object's `id` has an entry in `memo`, the value from `memo` is
    returned.

    If the object has a method `__lift__`, it is used to get the promise.
    `__lift__` should take one additional argument for the `memo` dictionary,
    entirely analogous to :func:`deepcopy`.

    If the object is an instance of one of the basic container types (list,
    dictionary, tuple and set), we use the analogous function
    (:func:`make_list`, :func:`make_dict`, :func:`make_tuple`, and
    :func:`make_set`) to promise their counterparts should these objects
    contain any promises.  First, we map all items in the container through
    :func:`lift`, then check the result for any promises.  Note that in the
    case of dictionaries, we lift all the items (i.e. the list of key/value
    tuples) and then construct a new dictionary.

    If the object is an instance of a subclass of any of the basic container
    types, the `__dict__` of the object is lifted as well as the object cast
    to its base type. We then use :func:`set_dict` to set the `__dict__` of
    the new promise. Again, if the object did not contain any promises,
    we return it without change.

    Otherwise, we lift the `__dict__` and create a promise of a new object of
    the same class as the input, using :func:`create_object`. This works fine
    for what we call *reasonable* objects. Since calling :func:`lift` is an
    explicit action, we do not require reasonable objects to be derived from
    :class:`Reasonable` as we do with serialisation, where such a default
    behaviour could lead to unexplicable bugs."""
    if memo is None:
        memo = {}

    if isinstance(obj, (PromisedObject, str, int, float, complex)):
        return obj

    if id(obj) in memo:
        return memo[id(obj)]

    if hasattr(obj, '__lift__'):
        rv = obj.__lift__(memo)
        memo[id(obj)] = rv
        return rv

    actions = {
        list:  (lambda x: x, make_list),
        dict:  (lambda x: list(x.items()), make_dict),
        tuple: (lambda x: x, make_tuple),
        set:   (lambda x: x, make_set)
    }

    if obj.__class__ in actions:
        items, construct = actions[obj.__class__]
        tmp = [lift(a, memo) for a in items(obj)]
        if any(isinstance(a, PromisedObject) for a in tmp):
            rv = construct(*tmp)
            memo[id(obj)] = rv
            return rv
        else:
            memo[id(obj)] = obj
            return obj

    subclass = next(filter(
        lambda x: issubclass(obj.__class__, x),
        actions.keys()), None)

    if subclass:
        members = lift(obj.__dict__, memo)
        internal = lift(subclass(obj), memo)

        if isinstance(internal, PromisedObject):
            internal = construct_object(obj.__class__, internal)
            rv = set_dict(internal, members)
        elif isinstance(members, PromisedObject):
            rv = set_dict(obj.__class__(internal), members)
        else:
            rv = obj

        memo[id(obj)] = rv
        return rv

    try:
        members = lift(obj.__dict__, memo)
        if isinstance(members, PromisedObject):
            rv = create_object(obj.__class__, members)
        else:
            rv = obj
    except AttributeError:
        memo[id(obj)] = obj
        return obj

    memo[id(obj)] = rv
    return rv



from pytest import raises
from noodles import (schedule, run_single)


class MyException(Exception):
    def __init__(self, msg):
        super(MyException, self).__init__(msg)


@schedule
def raises_my_exception():
    raise MyException("Error!")


def test_exception_00():
    with raises(MyException):
        wf = raises_my_exception()
        run_single(wf)



from noodles import schedule
from noodles.workflow import (
    invert_links, get_workflow, is_workflow, Empty, ArgumentAddress,
    ArgumentKind, is_node_ready)


@schedule
def value(a):
    return a


@schedule
def add(a, b):
    return a+b


def test_invert_links():
    A = value(1)
    B = value(2)
    C = add(A, B)

    C = get_workflow(C)
    A = get_workflow(A)
    B = get_workflow(B)

    assert is_workflow(C)
    assert C.nodes[C.root].bound_args.args == (Empty, Empty)
    assert (C.root, ArgumentAddress(ArgumentKind.regular, 'a', None)) \
        in C.links[A.root]
    assert (C.root, ArgumentAddress(ArgumentKind.regular, 'b', None)) \
        in C.links[B.root]

    deps = invert_links(C.links)
    assert deps == {
        A.root: {},
        B.root: {},
        C.root: {
            ArgumentAddress(ArgumentKind.regular, 'a', None): A.root,
            ArgumentAddress(ArgumentKind.regular, 'b', None): B.root}}


def test_is_node_ready():
    A = value(1)
    B = add(1, A)
    A = get_workflow(A)
    B = get_workflow(B)

    assert is_node_ready(A.nodes[A.root])
    assert not is_node_ready(B.nodes[B.root])



"""
Process backend
===============

Run jobs using a process backend.
"""

import sys
import uuid
from subprocess import Popen, PIPE
import threading
# import logging

import random
from ..workflow import get_workflow
# from ..logger import log
from .scheduler import Scheduler
# from .protect import CatchExceptions
from .hybrid import hybrid_threaded_worker

from ..lib import (pull, push, Connection, object_name, EndOfQueue, FlushQueue)
from .messages import (EndOfWork)

from .remote.io import (JSONObjectReader, JSONObjectWriter)


def process_worker(registry, verbose=False, jobdirs=False,
                   init=None, finish=None, status=True):
    """Process worker"""
    name = "process-" + str(uuid.uuid4())

    cmd = [sys.prefix + "/bin/python", "-m", "noodles.pilot_job",
           "-name", name, "-registry", object_name(registry)]
    if verbose:
        cmd.append("-verbose")
    if jobdirs:
        cmd.append("-jobdirs")
    if not status:
        cmd.append("-nostatus")
    if init:
        cmd.extend(["-init", object_name(init)])
    if finish:
        cmd.extend(["-finish", object_name(finish)])

    remote = Popen(
        cmd,
        stdin=PIPE, stdout=PIPE, stderr=PIPE, universal_newlines=True)

    def read_stderr():
        """Read stderr of remote process and sends lines to logger."""
        for line in remote.stderr:
            print(name + ": " + line.rstrip())

    stderr_reader_thread = threading.Thread(target=read_stderr, daemon=True)
    stderr_reader_thread.start()

    @push
    def send_job():
        """Coroutine, sends jobs to remote worker over standard input."""
        reg = registry()

        sink = JSONObjectWriter(reg, remote.stdin)

        while True:
            msg = yield
            if msg is EndOfQueue:
                try:
                    sink.send(EndOfWork)
                except StopIteration:
                    pass

                remote.wait()
                stderr_reader_thread.join()

                return

            if msg is FlushQueue:
                continue

            sink.send(msg)

    @pull
    def get_result():
        """Generator, reading results from process standard output."""
        reg = registry()
        yield from JSONObjectReader(reg, remote.stdout)

    return Connection(get_result, send_job)


def run_process(workflow, *, n_processes, registry,
                verbose=False, jobdirs=False,
                init=None, finish=None, deref=False):
    """Run the workflow using a number of new python processes. Use this
    runner to test the workflow in a situation where data serial
    is needed.

    :param workflow:
        The workflow.
    :type workflow: `Workflow` or `PromisedObject`

    :param n_processes:
        Number of processes to start.

    :param registry:
        The serial registry.

    :param verbose:
        Request verbose output on worker side

    :param jobdirs:
        Create a new directory for each job to prevent filename collision.(NYI)

    :param init:
        An init function that needs to be run in each process before other jobs
        can be run. This should be a scheduled function returning True on
        success.

    :param finish:
        A function that wraps up when the worker closes down.

    :param deref:
        Set this to True to pass the result through one more encoding and
        decoding step with object derefencing turned on.
    :type deref: bool

    :returns: the result of evaluating the workflow
    :rtype: any
    """
    workers = {}
    for i in range(n_processes):
        new_worker = process_worker(registry, verbose, jobdirs, init, finish)
        workers['worker {0:2}'.format(i)] = new_worker

    worker_names = list(workers.keys())

    def random_selector(_):
        """Selects a worker to send a job to at random."""
        return random.choice(worker_names)

    master_worker = hybrid_threaded_worker(random_selector, workers)
    result = Scheduler().run(master_worker, get_workflow(workflow))

    for worker in workers.values():
        try:
            worker.sink().send(EndOfQueue)
        except StopIteration:
            pass

#        w.aux.join()

    if deref:
        return registry().dereference(result, host='localhost')
    else:
        return result
